@@ -290,6 +290,10 @@ def delete_workflow_executions(**kwargs):
     IMPL.delete_workflow_executions(**kwargs)
 
 
+def update_workflow_execution_state(**kwargs):
+    return IMPL.update_workflow_execution_state(**kwargs)
+
+
 # Tasks executions.
 
 def get_task_execution(id):
@@ -344,6 +348,10 @@ def delete_task_executions(**kwargs):
     return IMPL.delete_task_executions(**kwargs)
 
 
+def update_task_execution_state(**kwargs):
+    return IMPL.update_task_execution_state(**kwargs)
+
+
 # Delayed calls.
 
 def get_delayed_calls_to_start(time):
@@ -393,12 +401,20 @@ def get_next_cron_triggers(time):
     return IMPL.get_next_cron_triggers(time)
 
 
-def get_executions_to_clean(expiration_time, limit=None,
-                            max_finished_executions=None, columns=()):
-    return IMPL.get_executions_to_clean(
+def get_expired_executions(expiration_time, limit=None, columns=(),
+                           session=None):
+    return IMPL.get_expired_executions(
         expiration_time,
         limit,
+        columns
+    )
+
+
+def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
+                               session=None):
+    return IMPL.get_superfluous_executions(
         max_finished_executions,
+        limit,
         columns
     )
 
