@@ -248,7 +248,11 @@ class PythonAction(Action):
             action_ex_id=action_ex_id
         )
 
-        action_queue.schedule(self.action_ex, self.action_def, target)
+        action_queue.schedule_run_action(
+            self.action_ex,
+            self.action_def,
+            target
+        )
 
     @profiler.trace('action-run', hide_args=True)
     def run(self, input_dict, target, index=0, desc='', save=True,
@@ -393,14 +397,15 @@ class AdHocAction(PythonAction):
         )
 
     def _prepare_input(self, input_dict):
-        for k, v in self.action_spec.get_input().items():
-            if k not in input_dict or input_dict[k] is utils.NotDefined:
-                input_dict[k] = v
-
         base_input_dict = input_dict
 
         for action_def in self.adhoc_action_defs:
             action_spec = spec_parser.get_action_spec(action_def.spec)
+            for k, v in action_spec.get_input().items():
+                if (k not in base_input_dict or
+                        base_input_dict[k] is utils.NotDefined):
+                    base_input_dict[k] = v
+
             base_input_expr = action_spec.get_base_input()
 
             if base_input_expr:
@@ -421,18 +426,20 @@ class AdHocAction(PythonAction):
     def _prepare_output(self, result):
         # In case of error, we don't transform a result.
         if not result.is_error():
-            adhoc_action_spec = spec_parser.get_action_spec(
-                self.adhoc_action_def.spec
-            )
-
-            transformer = adhoc_action_spec.get_output()
-
-            if transformer is not None:
-                result = ml_actions.Result(
-                    data=expr.evaluate_recursively(transformer, result.data),
-                    error=result.error
+            for action_def in reversed(self.adhoc_action_defs):
+                adhoc_action_spec = spec_parser.get_action_spec(
+                    action_def.spec
                 )
 
+                transformer = adhoc_action_spec.get_output()
+
+                if transformer is not None:
+                    result = ml_actions.Result(
+                        data=expr.evaluate_recursively(transformer,
+                                                       result.data),
+                        error=result.error
+                    )
+
         return result
 
     def _prepare_runtime_context(self, index, safe_rerun):
@@ -489,6 +496,10 @@ class AdHocAction(PythonAction):
 class WorkflowAction(Action):
     """Workflow action."""
 
+    def __init__(self, wf_name, **kwargs):
+        super(WorkflowAction, self).__init__(None, **kwargs)
+        self.wf_name = wf_name
+
     @profiler.trace('workflow-action-complete', hide_args=True)
     def complete(self, result):
         # No-op because in case of workflow result is already processed.
@@ -503,15 +514,11 @@ class WorkflowAction(Action):
             parent_wf_ex.id
         )
 
-        task_spec = spec_parser.get_task_spec(self.task_ex.spec)
-
-        wf_spec_name = task_spec.get_workflow_name()
-
         wf_def = engine_utils.resolve_workflow_definition(
             parent_wf_ex.workflow_name,
             parent_wf_spec.get_name(),
             namespace=parent_wf_ex.params['namespace'],
-            wf_spec_name=wf_spec_name
+            wf_spec_name=self.wf_name
         )
 
         wf_spec = spec_parser.get_workflow_spec_by_definition_id(
@@ -519,7 +526,13 @@ class WorkflowAction(Action):
             wf_def.updated_at
         )
 
+        # If the parent has a root_execution_id, it must be a sub-workflow. So
+        # we should propogate that ID down. Otherwise the parent must be the
+        # root execution and we should use the parents ID.
+        root_execution_id = parent_wf_ex.root_execution_id or parent_wf_ex.id
+
         wf_params = {
+            'root_execution_id': root_execution_id,
             'task_execution_id': self.task_ex.id,
             'index': index,
             'namespace': parent_wf_ex.params['namespace']
